@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  *
  * Authors: Jussi Kukkonen <jku@linux.intel.com>
  *          David Woodhouse <dwmw2@infradead.org>
@@ -40,16 +40,53 @@
 
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
+#include <glib-unix.h>
 
 #include <nm-vpn-plugin-utils.h>
 
+#include <gnome-keyring.h>
+
 #include "src/nm-openconnect-service.h"
 
 #include "openconnect.h"
 
+#if OPENCONNECT_API_VERSION_MAJOR == 1
+#define openconnect_vpninfo_new openconnect_vpninfo_new_with_cbdata
+#define openconnect_init_ssl openconnect_init_openssl
+#endif
+
+#ifndef OPENCONNECT_CHECK_VER
+#define OPENCONNECT_CHECK_VER(x,y) 0
+#endif
+
+#if !OPENCONNECT_CHECK_VER(1,5)
+#define OPENCONNECT_X509 X509
+#define OPENCONNECT_OPENSSL
+#endif
+
+#if !OPENCONNECT_CHECK_VER(2,1)
+#define openconnect_set_stoken_mode(...) -EOPNOTSUPP
+#endif
+
+#ifdef OPENCONNECT_OPENSSL
 #include <openssl/ssl.h>
 #include <openssl/bio.h>
 #include <openssl/ui.h>
+#endif
+
+static const GnomeKeyringPasswordSchema OPENCONNECT_SCHEMA_DEF = {
+  GNOME_KEYRING_ITEM_GENERIC_SECRET,
+  {
+    {"vpn_uuid", GNOME_KEYRING_ATTRIBUTE_TYPE_STRING},
+    {"auth_id", GNOME_KEYRING_ATTRIBUTE_TYPE_STRING},
+    {"label", GNOME_KEYRING_ATTRIBUTE_TYPE_STRING},
+    {NULL, 0}
+  }
+};
+
+const GnomeKeyringPasswordSchema *OPENCONNECT_SCHEMA = &OPENCONNECT_SCHEMA_DEF;
+
+static void got_keyring_pw(GnomeKeyringResult result, const char *string, gpointer data);
 
 static char *lasthost;
 
@@ -74,11 +111,53 @@ struct gconf_key {
 	struct gconf_key *next;
 };
 
+/* This struct holds all information we need to add a password to
+ * gnome-keyring. It’s used in success_passwords. */
+struct keyring_password {
+	char *description;
+	char *password;
+	char *vpn_uuid;
+	char *auth_id;
+	char *label;
+};
+
+static void keyring_password_free(gpointer data);
+static void keyring_store_passwords(gpointer key, gpointer value, gpointer user_data);
+
+static void keyring_password_free(gpointer data)
+{
+	struct keyring_password *kp = (struct keyring_password*)data;
+	g_free(kp->description);
+	g_free(kp->password);
+	g_free(kp->vpn_uuid);
+	g_free(kp->auth_id);
+	g_free(kp->label);
+	g_free(kp);
+}
+
+static void keyring_store_passwords(gpointer key, gpointer value, gpointer user_data)
+{
+	struct keyring_password *kp = (struct keyring_password*)value;
+	gnome_keyring_store_password_sync (
+			OPENCONNECT_SCHEMA,
+			GNOME_KEYRING_DEFAULT,
+			kp->description,
+			kp->password,
+			"vpn_uuid", kp->vpn_uuid,
+			"auth_id", kp->auth_id,
+			"label", kp->label,
+			NULL
+			);
+}
+
+
 typedef struct auth_ui_data {
 	char *vpn_name;
+	char *vpn_uuid;
 	GHashTable *options;
 	GHashTable *secrets;
 	GHashTable *success_secrets;
+	GHashTable *success_passwords;
 	struct openconnect_info *vpninfo;
 	GtkWidget *dialog;
 	GtkWidget *combo;
@@ -94,6 +173,7 @@ typedef struct auth_ui_data {
 	int retval;
 	int cookie_retval;
 
+	int cancel_pipes[2];
 	gboolean cancelled; /* fully cancel the whole challenge-response series */
 	gboolean getting_cookie;
 
@@ -205,8 +285,12 @@ static void ssl_box_clear(auth_ui_data *ui_data)
 
 typedef struct ui_fragment_data {
 	GtkWidget *widget;
+	GtkWidget *entry;
+	gpointer find_request;
 	auth_ui_data *ui_data;
+#ifdef OPENCONNECT_OPENSSL
 	UI_STRING *uis;
+#endif
 	struct oc_form_opt *opt;
 	char *entry_text;
 	int grab_focus;
@@ -217,6 +301,7 @@ static void entry_activate_cb(GtkWidget *widget, auth_ui_data *ui_data)
 	gtk_dialog_response(GTK_DIALOG(ui_data->dialog), AUTH_DIALOG_RESPONSE_LOGIN);
 }
 
+#ifdef OPENCONNECT_OPENSSL
 static void do_check_visibility(ui_fragment_data *data, gboolean *visible)
 {
 	int min_len;
@@ -229,13 +314,14 @@ static void do_check_visibility(ui_fragment_data *data, gboolean *visible)
 	if (min_len && (!data->entry_text || strlen(data->entry_text) < min_len))
 		*visible = FALSE;
 }
-
+#endif
 static void evaluate_login_visibility(auth_ui_data *ui_data)
 {
 	gboolean visible = TRUE;
+#ifdef OPENCONNECT_OPENSSL
 	g_queue_foreach(ui_data->form_entries, (GFunc)do_check_visibility,
 			&visible);
-
+#endif
 	gtk_widget_set_sensitive (ui_data->login_button, visible);
 }
 
@@ -243,7 +329,9 @@ static void entry_changed(GtkEntry *entry, ui_fragment_data *data)
 {
 	g_free (data->entry_text);
 	data->entry_text = g_strdup(gtk_entry_get_text(entry));
+#ifdef OPENCONNECT_OPENSSL
 	evaluate_login_visibility(data->ui_data);
+#endif
 }
 
 static void do_override_label(ui_fragment_data *data, struct oc_choice *choice)
@@ -272,6 +360,7 @@ static void combo_changed(GtkComboBox *combo, ui_fragment_data *data)
 			&sopt->choices[entry]);
 }
 
+#ifdef OPENCONNECT_OPENSSL
 static gboolean ui_write_error (ui_fragment_data *data)
 {
 	ssl_box_add_error(data->ui_data, UI_get0_output_string(data->uis));
@@ -289,6 +378,7 @@ static gboolean ui_write_info (ui_fragment_data *data)
 
 	return FALSE;
 }
+#endif
 
 static gboolean ui_write_prompt (ui_fragment_data *data)
 {
@@ -297,10 +387,13 @@ static gboolean ui_write_prompt (ui_fragment_data *data)
 	int visible;
 	const char *label;
 
+#ifdef OPENCONNECT_OPENSSL
 	if (data->uis) {
 		label = UI_get0_output_string(data->uis);
 		visible = UI_get_input_flags(data->uis) & UI_INPUT_FLAG_ECHO;
-	} else {
+	} else 
+#endif
+	{
 		label = data->opt->label;
 		visible = (data->opt->type == OC_FORM_OPT_TEXT);
 	}
@@ -318,14 +411,20 @@ static gboolean ui_write_prompt (ui_fragment_data *data)
 
 	entry = gtk_entry_new();
 	gtk_box_pack_end(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
+	data->entry = entry;
 	if (!visible)
 		gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
 	if (data->entry_text)
 		gtk_entry_set_text(GTK_ENTRY(entry), data->entry_text);
+	/* If it's the first empty one, grab focus. Otherwise, if
+	   it's the first item of *any* kind, grab focus but don't
+	   admit it (so another empty entry can take focus_ */
 	if (!data->entry_text && !data->ui_data->form_grabbed) {
 		data->ui_data->form_grabbed = 1;
 		gtk_widget_grab_focus (entry);
-	}
+	} else if (g_queue_peek_tail(ui_data->form_entries) == data)
+		gtk_widget_grab_focus (entry);
+
 	g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(entry_changed), data);
 	g_signal_connect(G_OBJECT(entry), "activate", G_CALLBACK(entry_activate_cb), ui_data);
 
@@ -393,6 +492,7 @@ static gboolean ui_show (auth_ui_data *ui_data)
 	return FALSE;
 }
 
+#ifdef OPENCONNECT_OPENSSL
 /* runs in worker thread */
 static int ui_open(UI *ui)
 {
@@ -480,6 +580,9 @@ static int ui_flush(UI* ui)
 		if (data->entry_text) {
 			UI_set_result(ui, data->uis, data->entry_text);
 		}
+		if (data->find_request) {
+			gnome_keyring_cancel_request(data->find_request);
+		}
 		g_slice_free (ui_fragment_data, data);
 	}
 	ui_data->form_grabbed = 0;
@@ -509,6 +612,7 @@ static int init_openssl_ui(void)
 	UI_set_default_method(ui_method);
 	return 0;
 }
+#endif /* OPENCONNECT_OPENSSL */
 
 static char *find_form_answer(GHashTable *secrets, struct oc_auth_form *form,
 			      struct oc_form_opt *opt)
@@ -521,6 +625,29 @@ static char *find_form_answer(GHashTable *secrets, struct oc_auth_form *form,
 	return result;
 }
 
+/* Callback which is called when we got a reply from gnome-keyring for any
+ * password field. Updates the contents of the password field unless the user
+ * entered anything in the meantime. */
+static void got_keyring_pw(GnomeKeyringResult result, const char *string, gpointer userdata)
+{
+	ui_fragment_data *data = (ui_fragment_data*)userdata;
+	if (string != NULL) {
+		if (data->entry) {
+			if (!g_ascii_strcasecmp("",
+						gtk_entry_get_text(GTK_ENTRY(data->entry)))) {
+				gtk_entry_set_text(GTK_ENTRY(data->entry), string);
+				if (gtk_widget_has_focus(data->entry))
+					gtk_editable_select_region(GTK_EDITABLE(data->entry), 0, -1);
+			}
+		} else
+			data->entry_text = g_strdup (string);
+	}
+
+	/* zero the find request so that we don’t attempt to cancel it when
+	 * closing the dialog */
+	data->find_request = NULL;
+}
+
 /* This part for processing forms from openconnect directly, rather than
    through the SSL UI abstraction (which doesn't allow 'select' options) */
 
@@ -562,6 +689,19 @@ static gboolean ui_form (struct oc_auth_form *form)
 			if (opt->type != OC_FORM_OPT_PASSWORD)
 				data->entry_text = g_strdup (find_form_answer(ui_data->secrets,
 									      form, opt));
+			else {
+				data->find_request = gnome_keyring_find_password(
+						OPENCONNECT_SCHEMA,
+						got_keyring_pw,
+						data,
+						NULL,
+						"vpn_uuid", ui_data->vpn_uuid,
+						"auth_id", form->auth_id,
+						"label", data->opt->name,
+						NULL
+						);
+			}
+
 
 			ui_write_prompt(data);
 		} else if (opt->type == OC_FORM_OPT_SELECT) {
@@ -579,10 +719,9 @@ static gboolean ui_form (struct oc_auth_form *form)
 	return ui_show(ui_data);
 }
 
-static int nm_process_auth_form (struct openconnect_info *vpninfo,
-				 struct oc_auth_form *form)
+static int nm_process_auth_form (void *cbdata, struct oc_auth_form *form)
 {
-	auth_ui_data *ui_data = _ui_data; /* FIXME global */
+	auth_ui_data *ui_data = cbdata;
 	int response;
 
 	g_idle_add((GSourceFunc)ui_form, form);
@@ -609,6 +748,10 @@ static int nm_process_auth_form (struct openconnect_info *vpninfo,
 		while (!g_queue_is_empty (ui_data->form_entries)) {
 			ui_fragment_data *data;
 			data = g_queue_pop_tail (ui_data->form_entries);
+
+			if (data->find_request)
+				gnome_keyring_cancel_request(data->find_request);
+
 			if (data->entry_text) {
 				data->opt->value = g_strdup (data->entry_text);
 
@@ -619,12 +762,26 @@ static int nm_process_auth_form (struct openconnect_info *vpninfo,
 					g_hash_table_insert (ui_data->success_secrets,
 							     keyname, g_strdup (data->entry_text));
 				}
+
+				if (data->opt->type == OC_FORM_OPT_PASSWORD) {
+					/* store the password in gnome-keyring */
+					//int result;
+					struct keyring_password *kp = g_new(struct keyring_password, 1);
+					kp->description = g_strdup_printf(_("OpenConnect: %s: %s:%s"), ui_data->vpn_name, form->auth_id, data->opt->name);
+					kp->password = g_strdup(data->entry_text);
+					kp->vpn_uuid = g_strdup(ui_data->vpn_uuid);
+					kp->auth_id = g_strdup(form->auth_id);
+					kp->label = g_strdup(data->opt->name);
+
+					g_hash_table_insert (ui_data->success_passwords,
+							g_strdup(kp->description), kp);
+				}
 			}
 			g_slice_free (ui_fragment_data, data);
 		}
 	}
 
-
+	ui_data->form_grabbed = 0;
 	g_mutex_unlock(ui_data->form_mutex);
 	
 	/* -1 = cancel,
@@ -637,35 +794,47 @@ static int nm_process_auth_form (struct openconnect_info *vpninfo,
 static char* get_title(const char *vpn_name)
 {
 	if (vpn_name)
-		return g_strdup_printf("Connect to VPN '%s'", vpn_name);
+		return g_strdup_printf(_("Connect to VPN '%s'"), vpn_name);
 	else
-		return g_strdup("Connect to VPN");
+		return g_strdup(_("Connect to VPN"));
 }
 
 typedef struct cert_data {
 	auth_ui_data *ui_data;
-	X509 *peer_cert;
+	OPENCONNECT_X509 *peer_cert;
 	const char *reason;
 } cert_data;
 
+#if !OPENCONNECT_CHECK_VER(1,5)
+static char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
+					  OPENCONNECT_X509 *cert)
+{
+        BIO *bp = BIO_new(BIO_s_mem());
+        BUF_MEM *certinfo;
+        char zero = 0;
+        char *ret;
+
+        X509_print_ex(bp, cert, 0, 0);
+        BIO_write(bp, &zero, 1);
+        BIO_get_mem_ptr(bp, &certinfo);
+
+        ret = strdup(certinfo->data);
+        BIO_free(bp);
+
+        return ret;
+}
+#endif
 
 static gboolean user_validate_cert(cert_data *data)
 {
 	auth_ui_data *ui_data = _ui_data; /* FIXME global */
-	BIO *bp = BIO_new(BIO_s_mem());
 	char *title;
-	BUF_MEM *certinfo;
-	char zero = 0;
+	char *details;
 	GtkWidget *dlg, *text, *scroll;
 	GtkTextBuffer *buffer;
 	int result;
 
-	/* There are probably better ways to do this -- getting individual
-	   elements of the cert info and formatting it nicely in the dialog
-	   box. But this will do for now... */
-	X509_print_ex(bp, data->peer_cert, 0, 0);
-	BIO_write(bp, &zero, 1);
-	BIO_get_mem_ptr(bp, &certinfo);
+	details = openconnect_get_cert_details(ui_data->vpninfo, data->peer_cert);
 
 	title = get_title(data->ui_data->vpn_name);
 	dlg = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION,
@@ -689,7 +858,8 @@ static gboolean user_validate_cert(cert_data *data)
 
 	text = gtk_text_view_new();
 	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text));
-	gtk_text_buffer_set_text(buffer, certinfo->data, -1);
+	gtk_text_buffer_set_text(buffer, details, -1);
+	free(details);
 	gtk_text_view_set_editable(GTK_TEXT_VIEW(text), 0);
 	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text), FALSE);
 	gtk_container_add(GTK_CONTAINER(scroll), text);
@@ -697,7 +867,6 @@ static gboolean user_validate_cert(cert_data *data)
 
 	result = gtk_dialog_run(GTK_DIALOG(dlg));
 
-	BIO_free(bp);
 	gtk_widget_destroy(dlg);
 
 	g_mutex_lock (ui_data->form_mutex);
@@ -712,16 +881,16 @@ static gboolean user_validate_cert(cert_data *data)
 }
 
 /* runs in worker thread */
-static int validate_peer_cert(struct openconnect_info *vpninfo,
-			      X509 *peer_cert, const char *reason)
+static int validate_peer_cert(void *cbdata,
+			      OPENCONNECT_X509 *peer_cert, const char *reason)
 {
-	auth_ui_data *ui_data = _ui_data; /* FIXME global */
-	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
+	auth_ui_data *ui_data = cbdata;
+	char fingerprint[41];
 	char *certs_data;
 	int ret = 0;
 	cert_data *data;
 
-	ret = openconnect_get_cert_sha1(vpninfo, peer_cert, fingerprint);
+	ret = openconnect_get_cert_sha1(ui_data->vpninfo, peer_cert, fingerprint);
 	if (ret)
 		return ret;
 
@@ -785,6 +954,16 @@ static gboolean get_autoconnect(GHashTable *secrets)
 	return FALSE;
 }
 
+static gboolean get_save_passwords(GHashTable *secrets)
+{
+	char *save = g_hash_table_lookup (secrets, "save_passwords");
+
+	if (save && !strcmp(save, "yes"))
+		return TRUE;
+
+	return FALSE;
+}
+
 static int parse_xmlconfig(gchar *xmlconfig)
 {
 	xmlDocPtr xml_doc;
@@ -868,6 +1047,8 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 	char *csd_wrapper;
 	char *pem_passphrase_fsid;
 	char *cafile;
+	char *stoken_source;
+	char *stoken_string;
 
 	hostname = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_GATEWAY);
 	if (!hostname) {
@@ -893,23 +1074,20 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 
 	xmlconfig = g_hash_table_lookup (secrets, "xmlconfig");
 	if (xmlconfig) {
+		GChecksum *sha1;
 		gchar *config_str;
 		gsize config_len;
-		unsigned char sha1[SHA_DIGEST_LENGTH];
-		char sha1_text[SHA_DIGEST_LENGTH * 2];
-		EVP_MD_CTX c;
-		int i;
+		const char *sha1_text;
 
 		config_str = (gchar *)g_base64_decode (xmlconfig, &config_len);
 
-		EVP_MD_CTX_init (&c);
-		EVP_Digest (config_str, config_len, sha1, NULL, EVP_sha1(), NULL);
-		EVP_MD_CTX_cleanup (&c);
+		sha1 = g_checksum_new (G_CHECKSUM_SHA1);
+		g_checksum_update (sha1, (gpointer) config_str, config_len);
+		sha1_text = g_checksum_get_string(sha1);
 
-		for (i = 0; i < SHA_DIGEST_LENGTH; i++)
-			sprintf (&sha1_text[i*2], "%02x", sha1[i]);
-
-		openconnect_set_xmlsha1 (vpninfo, sha1_text, sizeof(sha1_text));
+		openconnect_set_xmlsha1 (vpninfo, (char *)sha1_text, strlen(sha1_text) + 1);
+		g_checksum_free(sha1);
+		
 		parse_xmlconfig (config_str);
 	}
 
@@ -917,17 +1095,18 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 	if (cafile)
 		openconnect_set_cafile(vpninfo, g_strdup (cafile));
 
-	csd = g_hash_table_lookup (options, "enable_csd_trojan");
+	csd = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_CSD_ENABLE);
 	if (csd && !strcmp(csd, "yes")) {
 		/* We're not running as root; we can't setuid(). */
-		csd_wrapper = g_hash_table_lookup (options, "csd_wrapper");
+		csd_wrapper = g_hash_table_lookup (options,
+						   NM_OPENCONNECT_KEY_CSD_WRAPPER);
 		if (csd_wrapper && !csd_wrapper[0])
 			csd_wrapper = NULL;
 
 		openconnect_setup_csd(vpninfo, getuid(), 1, g_strdup (csd_wrapper));
 	}
 
-	proxy = g_hash_table_lookup (options, "proxy");
+	proxy = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_PROXY);
 	if (proxy && proxy[0] && openconnect_set_http_proxy(vpninfo, g_strdup (proxy)))
 		return -EINVAL;
 
@@ -935,10 +1114,25 @@ static int get_config (GHashTable *options, GHashTable *secrets,
 	sslkey = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_PRIVKEY);
 	openconnect_set_client_cert (vpninfo, g_strdup (cert), g_strdup (sslkey));
 
-	pem_passphrase_fsid = g_hash_table_lookup (options, "pem_passphrase_fsid");
+	pem_passphrase_fsid = g_hash_table_lookup (options,
+						   NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
 	if (pem_passphrase_fsid && cert && !strcmp(pem_passphrase_fsid, "yes"))
 		openconnect_passphrase_from_fsid(vpninfo);
 
+	stoken_source = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_STOKEN_SOURCE);
+	stoken_string = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_STOKEN_STRING);
+	if (stoken_source) {
+		int ret = 0;
+
+		if (!strcmp(stoken_source, "manual") && stoken_string)
+			ret = openconnect_set_stoken_mode(vpninfo, 1, stoken_string);
+		else if (!strcmp(stoken_source, "stokenrc"))
+			ret = openconnect_set_stoken_mode(vpninfo, 1, NULL);
+
+		if (ret)
+			fprintf(stderr, "Failed to initialize stoken: %d\n", ret);
+	}
+
 	return 0;
 }
 
@@ -959,9 +1153,9 @@ static void populate_vpnhost_combo(auth_ui_data *ui_data)
 	}
 }
 
-static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen)
+static int write_new_config(void *cbdata, char *buf, int buflen)
 {
-	auth_ui_data *ui_data = _ui_data; /* FIXME global */
+	auth_ui_data *ui_data = cbdata;
 	g_hash_table_insert (ui_data->secrets, g_strdup ("xmlconfig"),
 			     g_base64_encode ((guchar *)buf, buflen));
 
@@ -981,6 +1175,41 @@ static void autocon_toggled(GtkWidget *widget)
 	g_hash_table_insert (ui_data->secrets, g_strdup ("autoconnect"), enabled);
 }
 
+/* gnome_keyring_delete_password() only deletes one matching password, so
+   keep doing it until it doesn't succeed. The ui_data is essentially
+   permanent anyway so no need to worry about its lifetime. */
+static void delete_next_password(GnomeKeyringResult result, gpointer data)
+{
+	auth_ui_data *ui_data = data;
+
+	if (result == GNOME_KEYRING_RESULT_OK) {
+		gnome_keyring_delete_password(OPENCONNECT_SCHEMA,
+					      delete_next_password,
+					      ui_data, NULL,
+					      "vpn_uuid", ui_data->vpn_uuid,
+					      NULL);
+	}		
+		
+}
+
+static void savepass_toggled(GtkWidget *widget)
+{
+	auth_ui_data *ui_data = _ui_data; /* FIXME global */
+	gchar *enabled;
+
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget)))
+		enabled = g_strdup ("yes");
+	else {
+		enabled = g_strdup ("no");
+		gnome_keyring_delete_password(OPENCONNECT_SCHEMA,
+					      delete_next_password,
+					      ui_data, NULL,
+					      "vpn_uuid", ui_data->vpn_uuid,
+					      NULL);
+	}
+	g_hash_table_insert (ui_data->secrets, g_strdup ("save_passwords"), enabled);
+}
+
 static void scroll_log(GtkTextBuffer *log, GtkTextView *view)
 {
 	GtkTextMark *mark;
@@ -1021,7 +1250,7 @@ static gboolean write_notice_real(char *message)
 }
 
 /* runs in worker thread */
-static void write_progress(struct openconnect_info *info, int level, const char *fmt, ...)
+static void write_progress(void *cbdata, int level, const char *fmt, ...)
 {
 	va_list args;
 	char *msg;
@@ -1059,6 +1288,7 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 	if (ui_data->cancelled) {
 		/* user has chosen a new host, start from beginning */
 		g_hash_table_remove_all (ui_data->success_secrets);
+		g_hash_table_remove_all (ui_data->success_passwords);
 		connect_host(ui_data);
 		return FALSE;
 	}
@@ -1070,11 +1300,11 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 						 GTK_STOCK_DIALOG_ERROR,
 						 GTK_ICON_SIZE_DIALOG);
 			gtk_widget_show_all(ui_data->ssl_box);
-			gtk_widget_set_sensitive(ui_data->cancel_button, TRUE);
+			gtk_widget_set_sensitive(ui_data->cancel_button, FALSE);
 		}
 		ui_data->retval = 1;
 	} else if (!ui_data->cookie_retval) {
-		X509 *cert;
+		OPENCONNECT_X509 *cert;
 		gchar *key, *value;
 
 		/* got cookie */
@@ -1099,11 +1329,16 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 		cert = openconnect_get_peer_cert (ui_data->vpninfo);
 		if (cert) {
 			key = g_strdup (NM_OPENCONNECT_KEY_GWCERT);
-			value = g_malloc0 (EVP_MAX_MD_SIZE * 2 + 1);
+			value = g_malloc0 (41);
 			openconnect_get_cert_sha1(ui_data->vpninfo, cert, value);
 			g_hash_table_insert (ui_data->secrets, key, value);
 		}
 
+		if (get_save_passwords(ui_data->secrets)) {
+			g_hash_table_foreach(ui_data->success_passwords,
+					     keyring_store_passwords,
+					     NULL);
+		}
 		ui_data->retval = 0;
 
 		gtk_main_quit();
@@ -1114,6 +1349,7 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 	}
 
 	g_hash_table_remove_all (ui_data->success_secrets);
+	g_hash_table_remove_all (ui_data->success_passwords);
 
 	return FALSE;
 }
@@ -1121,9 +1357,13 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 static gpointer obtain_cookie (auth_ui_data *ui_data)
 {
 	int ret;
+	char cancelbuf;
 
 	ret = openconnect_obtain_cookie(ui_data->vpninfo);
 
+	/* Suck out the poison */
+	while (read(ui_data->cancel_pipes[0], &cancelbuf, 1) == 1)
+		;
 	ui_data->cookie_retval = ret;
 	g_idle_add ((GSourceFunc)cookie_obtained, ui_data);
 
@@ -1136,6 +1376,7 @@ static void connect_host(auth_ui_data *ui_data)
 	vpnhost *host;
 	int i;
 	int host_nr;
+	char cancelbuf;
 
 	ui_data->cancelled = FALSE;
 	ui_data->getting_cookie = TRUE;
@@ -1146,7 +1387,9 @@ static void connect_host(auth_ui_data *ui_data)
 
 	ssl_box_clear(ui_data);
 	gtk_widget_show(ui_data->getting_form_label);
-
+	gtk_widget_set_sensitive (ui_data->cancel_button, TRUE);
+	while (read(ui_data->cancel_pipes[0], &cancelbuf, 1) == 1)
+		;
 	/* reset ssl context.
 	 * TODO: this is probably not the way to go... */
 	openconnect_reset_ssl(ui_data->vpninfo);
@@ -1200,6 +1443,10 @@ static void dialog_response (GtkDialog *dialog, int response, auth_ui_data *ui_d
 {
 	switch (response) {
 	case AUTH_DIALOG_RESPONSE_CANCEL:
+		if (write(ui_data->cancel_pipes[1], "x", 1) < 0) {
+			/* Pfft. Not a lot we can do about it */
+		}
+		/* Fall through... */
 	case AUTH_DIALOG_RESPONSE_LOGIN:
 		ssl_box_clear(ui_data);
 		if (ui_data->getting_cookie)
@@ -1231,7 +1478,7 @@ static void build_main_dialog(auth_ui_data *ui_data)
 {
 	char *title;
 	GtkWidget *vbox, *hbox, *label, *frame, *image, *frame_box;
-	GtkWidget *exp, *scrolled, *view, *autocon;
+	GtkWidget *exp, *scrolled, *view, *autocon, *save_pass;
 
 	gtk_window_set_default_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION);
 
@@ -1342,6 +1589,14 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	gtk_widget_set_sensitive (ui_data->cancel_button, FALSE);
 	gtk_widget_show(ui_data->cancel_button);
 
+	save_pass = gtk_check_button_new_with_label(_("Save passwords"));
+	gtk_box_pack_start(GTK_BOX(hbox), save_pass, FALSE, FALSE, 0);
+	if (get_save_passwords (ui_data->secrets))
+		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(save_pass), 1);
+	g_signal_connect(save_pass, "toggled", G_CALLBACK(savepass_toggled), NULL);
+	gtk_widget_show(save_pass);
+
+
 	exp = gtk_expander_new(_("Log"));
 	gtk_box_pack_end(GTK_BOX(vbox), exp, FALSE, FALSE, 0);
 	gtk_widget_show(exp);
@@ -1367,7 +1622,7 @@ static void build_main_dialog(auth_ui_data *ui_data)
 	g_signal_connect(ui_data->log, "changed", G_CALLBACK(scroll_log), view);
 }
 
-static auth_ui_data *init_ui_data (char *vpn_name, GHashTable *options, GHashTable *secrets)
+static auth_ui_data *init_ui_data (char *vpn_name, GHashTable *options, GHashTable *secrets, char *vpn_uuid)
 {
 	auth_ui_data *ui_data;
 
@@ -1391,14 +1646,31 @@ static auth_ui_data *init_ui_data (char *vpn_name, GHashTable *options, GHashTab
 	ui_data->cert_response_changed = g_cond_new();
 #endif
 	ui_data->vpn_name = vpn_name;
+	ui_data->vpn_uuid = vpn_uuid;
 	ui_data->options = options;
 	ui_data->secrets = secrets;
 	ui_data->success_secrets = g_hash_table_new_full (g_str_hash, g_str_equal,
 							  g_free, g_free);
+	ui_data->success_passwords = g_hash_table_new_full (g_str_hash, g_str_equal,
+							  g_free, keyring_password_free);
+	if (pipe(ui_data->cancel_pipes)) {
+		/* This should never happen, and the world is probably about
+		   to come crashing down around our ears. But attempt to cope
+		   by just disabling the cancellation support... */
+		ui_data->cancel_pipes[0] = -1;
+		ui_data->cancel_pipes[1] = -1;
+	}
+	g_unix_set_fd_nonblocking(ui_data->cancel_pipes[0], TRUE, NULL);
+	g_unix_set_fd_nonblocking(ui_data->cancel_pipes[1], TRUE, NULL);
 
 	ui_data->vpninfo = (void *)openconnect_vpninfo_new("OpenConnect VPN Agent (NetworkManager)",
-						   validate_peer_cert, write_new_config,
-						   nm_process_auth_form, write_progress);
+							   validate_peer_cert, write_new_config,
+							   nm_process_auth_form, write_progress,
+							   ui_data);
+
+#if OPENCONNECT_CHECK_VER(1,4)
+	openconnect_set_cancel_fd (ui_data->vpninfo, ui_data->cancel_pipes[0]);
+#endif  
 
 #if 0
 	ui_data->vpninfo->proxy_factory = px_proxy_factory_new();
@@ -1407,6 +1679,30 @@ static auth_ui_data *init_ui_data (char *vpn_name, GHashTable *options, GHashTab
 	return ui_data;
 }
 
+static void wait_for_quit (void)
+{
+	GString *str;
+	char c;
+	ssize_t n;
+	time_t start;
+
+	str = g_string_sized_new (10);
+	start = time (NULL);
+	do {
+		errno = 0;
+		n = read (0, &c, 1);
+		if (n == 0 || (n < 0 && errno == EAGAIN))
+			g_usleep (G_USEC_PER_SEC / 10);
+		else if (n == 1) {
+			g_string_append_c (str, c);
+			if (strstr (str->str, "QUIT") || (str->len > 10))
+				break;
+		} else
+			break;
+	} while (time (NULL) < start + 20);
+	g_string_free (str, TRUE);
+}
+
 static struct option long_options[] = {
 	{"reprompt", 0, 0, 'r'},
 	{"uuid", 1, 0, 'u'},
@@ -1486,15 +1782,17 @@ int main (int argc, char **argv)
 #endif
 	gtk_init(0, NULL);
 
-	_ui_data = init_ui_data(vpn_name, options, secrets);
+	_ui_data = init_ui_data(vpn_name, options, secrets, vpn_uuid);
 	if (get_config(options, secrets, _ui_data->vpninfo)) {
 		fprintf(stderr, "Failed to find VPN UUID %s in gconf\n", vpn_uuid);
 		return 1;
 	}
 	build_main_dialog(_ui_data);
 
+#ifdef OPENCONNECT_OPENSSL
 	init_openssl_ui();
-	openconnect_init_openssl();
+#endif
+	openconnect_init_ssl();
 
 	if (get_autoconnect (secrets))
 		queue_connect_host(_ui_data);
@@ -1511,5 +1809,7 @@ int main (int argc, char **argv)
 	printf("\n\n");
 	fflush(stdout);
 
+	wait_for_quit ();
+
 	return _ui_data->retval;
 }
